package org.bdware.server.irp;

import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import com.sun.net.httpserver.HttpServer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.util.JsonUtil;
import org.rocksdb.Options;
import org.rocksdb.RocksDB;
import org.rocksdb.RocksDBException;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetSocketAddress;
import java.net.URLDecoder;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

public class LocalLHSProxy {
    private static final Logger LOGGER = LogManager.getLogger(LocalLHSProxy.class);
    public static String PREFIX = "86.5000.470/";
    public static int port = 18007;
    public static boolean enabled = false;
    public static RocksDB db;
    static RegisterHandler registerHandler;

    static {
        RocksDB.loadLibrary();
    }

    public static void main(String[] args) throws Exception {
        start();
    }

    public static void start() throws IOException, RocksDBException {
        String dbPath = "./handleRecords/";
        Options rockopts = new Options().setCreateIfMissing(true);
        rockopts.useFixedLengthPrefixExtractor(15);
        LOGGER.info("actual rocksdb path: " + new File(dbPath).getAbsolutePath());
        File lock = new File(dbPath, "LOCK");
        if (lock.exists()) {
            LOGGER.trace("remove file " + lock.getAbsolutePath() + ": " + lock.delete());
        }
        db = RocksDB.open(rockopts, dbPath);

        registerHandler = new RegisterHandler(db);

        LOGGER.info("listen to " + port);
        HttpServer server = HttpServer.create(new InetSocketAddress(port), 0);

        server.createContext("/test", new TestHandler());
        server.createContext("/resolve", new ResolveHandler());
        server.createContext("/local/", registerHandler);
        server.createContext("/view", new ViewHandler(db));
        server.start();
    }

    private static Map<String, String> formData2Dic(String formData) {
        Map<String, String> result = new HashMap<>();
        if (formData == null || formData.trim().length() == 0) {
            return result;
        }
        final String[] items = formData.split("&");
        Arrays.stream(items)
                .forEach(
                        item -> {
                            final String[] keyAndVal = item.split("=");
                            if (keyAndVal.length == 2) {
                                try {
                                    final String key = URLDecoder.decode(keyAndVal[0], "utf8");
                                    final String val = URLDecoder.decode(keyAndVal[1], "utf8");
                                    result.put(key, val);
                                } catch (UnsupportedEncodingException e) {
                                    e.printStackTrace();
                                }
                            }
                        });
        return result;
    }

    static class TestHandler implements HttpHandler {
        @Override
        public void handle(HttpExchange exchange) {
            String response = "hello world";
            try {
                exchange.sendResponseHeaders(200, 0);
                OutputStream os = exchange.getResponseBody();
                os.write(response.getBytes());
                os.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    static class ResolveHandler implements HttpHandler {
        @Override
        public void handle(HttpExchange exchange) {
            try {
                // 获得查询字符串(get)
                String queryString = exchange.getRequestURI().getQuery();
                Map<String, String> queryStringInfo = formData2Dic(queryString);
                String id = queryStringInfo.get("identifier");
                Map<String, String> respMap;
                respMap = registerHandler.handleResolve(queryStringInfo);
                respMap.forEach((k, v) -> {
                    LOGGER.debug(k + "->" + v);
                });
                exchange.getResponseHeaders().add("Access-Control-Allow-Origin", "*");
                exchange.sendResponseHeaders(200, 0);
                OutputStream os = exchange.getResponseBody();
                os.write(JsonUtil.toJson(respMap).getBytes());
                os.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}
