package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufInputStream;
import io.netty.buffer.Unpooled;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.NodeCenterServer;
import org.bdware.server.action.Action;
import org.bdware.server.action.ActionExecutor;

import java.io.ByteArrayOutputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class NodeCenterFrameHandler extends SimpleChannelInboundHandler<Object> {
    private static final Logger LOGGER = LogManager.getLogger(LogActions.class);
    static ExecutorService executorService = Executors.newFixedThreadPool(10);
    public String pubKey;
    NodeCenterActions actions;
    MasterActions masterActions;
    ActionExecutor<ResultCallback, JsonObject> ae;
    ChannelHandlerContext ctx;

    public NodeCenterFrameHandler() {
        this.actions = new NodeCenterActions(this);
        MetaIndexAction.controller = this;
        // TODO 添加那个UnitAction.
        ae =
                new ActionExecutor<ResultCallback, JsonObject>(
                        executorService, actions, new MetaIndexAction()) {
                    @Override
                    public boolean checkPermission(
                            Action a, final JsonObject args, long permission) {
                        long val = a.userPermission();
                        boolean flag;
                        String status = "refuse";
                        String action = args.get("action").getAsString();
                        if (val == 0) {
                            flag = true;
                            status = "accept";
                        } else if ((permission & val) == val) {
                            flag = true;
                            status = "accept";
                        } else {
                            flag = false;
                        }
                        if (!action.contains("checkAlive"))
                            NodeCenterServer.nodeTcpLogDB.put(
                                    action,
                                    "{\"action\":\""
                                            + action
                                            + "\",\"pubKey\":\""
                                            + pubKey
                                            + "\",\"status\":\""
                                            + status
                                            + "\",\"date\":"
                                            + System.currentTimeMillis()
                                            + "}");
                        return flag;
                    }
                };
        LOGGER.info("create NodeCenterFrameHandler instance succeed!");
    }

    public void setPermission(long l) {
        ae.permission = l;
    }

    @Override
    protected void channelRead0(ChannelHandlerContext ctx, Object frame) {
        this.ctx = ctx;
        JsonObject map;
        try {
            map =
                    JsonParser.parseReader(
                                    new InputStreamReader(new ByteBufInputStream((ByteBuf) frame)))
                            .getAsJsonObject();

        } catch (Exception e) {
            LOGGER.error("can't handle msg! " + e.getMessage());
            return;
        }
        Response response;
        try {
            final String action = map.get("action").getAsString();
            // System.out.println("[NodeCenterFramHandler] handle:" + action);
            ae.handle(
                    action,
                    map,
                    new ResultCallback() {
                        @Override
                        public void onResult(String ret) {
                            sendMsg(ret);
                        }
                    });
        } catch (IllegalArgumentException e) {
            response = new Response();
            response.action = "onException";
            response.data = e.getMessage();
            LOGGER.error(JsonUtil.toJson(response));
        } catch (Exception e) {
            e.printStackTrace();
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            response = new Response();
            response.action = "onException";
            String[] strs = bo.toString().split("\n");
            StringBuilder ret = new StringBuilder();
            int count = 0;
            for (String s : strs) {
                if (s.contains("sun.reflect")
                        || s.contains("java.lang.reflect")
                        || s.contains("org.apache")
                        || s.contains("java.util")
                        || s.contains("java.lang")
                        || s.contains("io.netty")) {
                    continue;
                }
                ret.append(s);
                ret.append("\n");
                if (count++ > 10) {
                    break;
                }
            }
            response.data = ret.toString();
            LOGGER.error(JsonUtil.toJson(response));
        }
    }

    public void sendMsg(String json) {
        if (ctx != null) {
            // System.out.println("[NodeCenterFrame send] TID:" + Thread.currentThread().getId() + "
            // isOpen:"
            //		+ ctx.channel().isOpen() + " isActive:" + ctx.channel().isActive() + " -->" + json);
            ByteBuf buf = Unpooled.wrappedBuffer(json.getBytes());
            ctx.channel().writeAndFlush(buf);
        }
    }

    public boolean isOpen() {
        return ctx.channel().isOpen();
    }

    static class Response {
        public String cid;
        public String analysisResult;
        String responseID;
        String action;
        Object data;
        long executeTime;
    }
}
