package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import io.netty.handler.codec.http.websocketx.WebSocketFrame;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.NodeCenterServer;
import org.bdware.server.action.Action;
import org.bdware.server.action.ActionExecutor;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class NodeCenterWSFrameHandler extends SimpleChannelInboundHandler<WebSocketFrame> {
    public static ExecutorService executorService = Executors.newFixedThreadPool(10);
    private static final Logger LOGGER = LogManager.getLogger(NodeCenterWSFrameHandler.class);
    StringBuilder dataCache = new StringBuilder();
    ActionExecutor<ResultCallback, JsonObject> ae;
    private ChannelHandlerContext ctx;
    private NCManagerAction managerAction;
    private LogActions logActions;
    private UnitActions unitActions;
    public FileActions fileActions;

    public NodeCenterWSFrameHandler() {
        managerAction = new NCManagerAction(this);
        logActions = new LogActions(managerAction);
        unitActions = new UnitActions(managerAction);
        fileActions = new FileActions(this);
        ae =
                new ActionExecutor<ResultCallback, JsonObject>(
                        executorService, managerAction, logActions, unitActions, fileActions, new MetaIndexAction(), new TracingAction()) {
                    @Override
                    public boolean checkPermission(Action a, JsonObject arg, long permission) {
                        // Permission userPermission = a.userPermission();
                        // long val=userPermission.getValue();
                        long val = a.userPermission();
                        String pubKey = managerAction.pubKey;
                        String action = arg.get("action").getAsString();
                        LOGGER.debug("permission" + permission);
                        LOGGER.debug("userPermission" + val);
                        NodeCenterServer.nodeHttpLogDB.put(
                                action,
                                "{\"action\":\""
                                        + action
                                        + "\",\"pubKey\":\""
                                        + pubKey
                                        + "\",\"date\":"
                                        + System.currentTimeMillis()
                                        + "}");

                        //                        if (val == 0) return true;
                        //                        return true;
                        if ((permission & val) == val) {
                            return true;
                        } else return false;
                    }
                };
    }

    public ActionExecutor getAE() {
        return ae;
    }

    public String getPubKey() {
        return managerAction.pubKey;
    }

    @Override
    protected void channelRead0(final ChannelHandlerContext ctx, WebSocketFrame frame)
            throws Exception {
        // ping and pong frames already handled
        if (frame instanceof TextWebSocketFrame) {
            // Send the uppercase string back.
            this.ctx = ctx;
            String request = ((TextWebSocketFrame) frame).text();
            LOGGER.debug("[Websocket receive]" + request);
            JsonObject map = null;
            try {
                map = new JsonParser().parse(request).getAsJsonObject();
            } catch (Exception e) {
                Response response = new Response();
                response.action = "onException";
                response.data = "msg is not JSON format";
                ctx.channel().writeAndFlush(new TextWebSocketFrame(JsonUtil.toJson(response)));
                return;
            }
            Response response = new Response();
            response.action = "sendNextSegment";
            try {
                boolean isSegment = false;
                if (map.has("isSegment")) isSegment = map.get("isSegment").getAsBoolean();
                if (isSegment) {
                    dataCache.append(map.get("data").getAsString());
                    response = new Response();
                    response.action = "sendNextSegment";
                    ctx.channel()
                            .writeAndFlush(new TextWebSocketFrame(JsonUtil.toJson(response)));
                    return;
                } else {
                    if (dataCache.length() > 0) {
                        dataCache.append(map.get("data").getAsString());
                        String dataStr = dataCache.toString();
                        LOGGER.debug("[WebSocketFrame] DataString:" + dataStr);
                        map = new JsonParser().parse(dataStr).getAsJsonObject();
                        dataCache = new StringBuilder();
                    }
                }
                String action = map.get("action").getAsString();
                final JsonObject jmap = map;
                ae.handle(
                        action,
                        map,
                        new ResultCallback() {
                            @Override
                            public void onResult(Map jo) {
                                if (jmap.has("requestID")) {
                                    jo.put("responseID", jmap.get("requestID"));
                                }
                                onResult(JsonUtil.toJson(jo));
                            }

                            @Override
                            public void onResult(JsonObject jo) {
                                if (jmap.has("requestID")) {
                                    jo.add("responseID", jmap.get("requestID"));
                                }
                                onResult(jo.toString());
                            }

                            @Override
                            public void onResult(String ret) {
                                if (ret != null)
                                    ctx.channel().writeAndFlush(new TextWebSocketFrame(ret));
                            }
                        });
            } catch (IllegalArgumentException e) {
                response = new Response();
                response.action = "onException";
                response.data = e.getMessage();
                ctx.channel().writeAndFlush(new TextWebSocketFrame((JsonUtil.toJson(response))));
                return;
            } catch (Exception e) {
                ByteArrayOutputStream bo = new ByteArrayOutputStream();
                e.printStackTrace();
                e.printStackTrace(new PrintStream(bo));
                response = new Response();
                response.action = "onException";
                String[] strs = bo.toString().split("\n");
                String ret = "";
                int count = 0;
                for (String s : strs) {
                    if (s.contains("sun.reflect")) continue;
                    if (s.contains("java.lang.reflect")) continue;
                    if (s.contains("org.apache")) continue;
                    if (s.contains("java.util")) continue;
                    if (s.contains("java.lang")) continue;
                    if (s.contains("io.netty")) continue;
                    ret += s;
                    ret += "\n";
                    if (count++ > 5) break;
                }
                response.data = ret;
                ctx.channel().writeAndFlush(new TextWebSocketFrame(JsonUtil.toJson(response)));
            }

        } else {
            Response response = new Response();
            response.action = "onException";
            response.data = "unsupported frame";
            ctx.channel().writeAndFlush(new TextWebSocketFrame((JsonUtil.toJson(response))));
            return;
        }
    }

    public void sendMsg(String json) {
        LOGGER.debug("[WebSocketFrameHandler] sendMsg:" + json);
        ctx.channel().writeAndFlush(new TextWebSocketFrame(json));
    }

    public void setPermission(long i) {
        ae.permission = i;
    }
}
