package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.bean.ContractDesp;
import org.bdware.sc.bean.ContractExecType;
import org.bdware.sc.bean.OtherNCInfo;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.NodeCenterServer;
import org.bdware.server.ws.DelimiterCodec;
import org.zz.gmhelper.SM2KeyPair;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

public class OtherNCProxy {
    private static final Logger LOGGER = LogManager.getLogger(OtherNCProxy.class);
    public static OtherNCProxy instance = new OtherNCProxy();
    private final TreeMap<String, List<OtherNCInfo>> otherNodeCenterInfos = new TreeMap<>(); //key为NC节点的pubkey(非一致的)
    public Map<String, NCConnector> connList = new ConcurrentHashMap<>();  //连接别的NC key为ip:port
    public Set<String> ncAdds = ConcurrentHashMap.newKeySet();  //其他NC的ip:port
    public SM2KeyPair sm2;  //NC证书

    private OtherNCProxy() {
    }

    // 从DB中加载OtherNC和NCFile的信息
    public void init() {
        LOGGER.debug("[OtherNCProxy] init : ");

        loadOtherNCs();
        LOGGER.debug("load other ncs");

        loadNCFile();
        LOGGER.debug("load nc file");

        periodUpdate();
        LOGGER.debug("period update");

        LOGGER.info("OtherNCProxy init done!");
    }

    public void periodUpdate() {
        NodeCenterServer.scheduledThreadPool.scheduleWithFixedDelay(
                () -> {
                },
                0,
                30,
                TimeUnit.SECONDS);
    }

    public void loadOtherNCs() {
        LOGGER.debug("loadOtherNCs : ");

        String res = getOtherNCs();
        if (res != null && !res.equals("")) {
            String[] strArr = res.split(";");
            ncAdds.clear();
            ncAdds.addAll(Arrays.asList(strArr));

            LOGGER.debug("now ncAdds contains: ");
            for (String s : strArr) {
                LOGGER.debug(s);
            }

            if (ncAdds.isEmpty()) {
                KeyValueDBUtil.instance.setValue(NCTables.OtherNCs.toString(), "address", "");
            }
        } else {
            ncAdds.clear();
        }
    }

    // 配置一致的NC证书
    public void loadNCFile() {
        LOGGER.debug("loadNCFile");

        String fileAdd = getNCFile();
        if (fileAdd != null && !fileAdd.isEmpty()) {
            LOGGER.debug("NCFile path =  " + fileAdd);

            File file = new File(fileAdd);
            if (!file.exists()) {
                LOGGER.warn("file not exists!");
                KeyValueDBUtil.instance.setValue(NCTables.NCFile.toString(), "filePath", "");
                return;
            }

            String keyPairStr = null;
            try {
                FileReader fr = new FileReader(file);
                BufferedReader br = new BufferedReader(fr);
                String s;
                while (null != (s = br.readLine())) {
                    keyPairStr = s;
                }
                fr.close();
                br.close();
            } catch (IOException e) {
                e.printStackTrace();
            }

            if (keyPairStr == null) {
                LOGGER.warn("import ket pair form NCFile is null!");
                return;
            }
            LOGGER.info("import ket pair from NCFile : " + keyPairStr);
            sm2 = SM2KeyPair.fromJson(keyPairStr);
        } else {
            sm2 = null;
        }
    }

    public String getOtherNCs() {
        return KeyValueDBUtil.instance.getValue(NCTables.OtherNCs.toString(), "address");
    }

    public String getNCFile() {
        return KeyValueDBUtil.instance.getValue(NCTables.NCFile.toString(), "filePath");
    }

    //配置其他NC
    public String setOtherNCs(String add) {
        LOGGER.debug("setOtherNC: " + add);

        KeyValueDBUtil.instance.setValue(NCTables.OtherNCs.toString(), "address", add);
        loadOtherNCs();
        return getOtherNCs();
    }

    //配置NC文件
    public String setNCFile(String fileName) {
        LOGGER.debug("setNCFIle: " + fileName);

        KeyValueDBUtil.instance.setValue(NCTables.NCFile.toString(), "filePath", fileName);
        loadNCFile();
        return getNCFile();
    }

    //本地更新
    public void updateOtherNCInfo(String nodePubkey, List<OtherNCInfo> list) {
        LOGGER.debug("updateOtherNCInfo: ");

        synchronized (otherNodeCenterInfos) {
            otherNodeCenterInfos.put(nodePubkey, list);
            LOGGER.debug("update route  nodePubkey=" + nodePubkey + "  list=" + JsonUtil.toJson(list));
        }
    }

    //通过合约id或name在别的nc上查询信息
    public String search(String id) {
        LOGGER.debug("search : " + id);
        synchronized (otherNodeCenterInfos) {
            for (String pubKey : otherNodeCenterInfos.keySet()) {
                List<OtherNCInfo> list = otherNodeCenterInfos.get(pubKey);
                for (OtherNCInfo info : list) {
                    if (info.contractID.equals(id) || info.contractName.equals(id)) {
                        LOGGER.debug("[OtherNCProxy] find contract " + id + "pubKey is " + pubKey + " master is " + info.master + " from other NC route info");
                        return pubKey + ";" + info.master;
                    }
                }
            }
            return null;
        }
    }

    //向其他NC发起同步
    public void requestUpdate() {
        LOGGER.debug("sendUpdate: ");
        for (String add : instance.ncAdds) {
            try {
                instance.connectToNC(add);
            } catch (Exception e) {
                LOGGER.error("NC " + add + " can not be connected! " + e.getMessage());
                connList.get(add).close();
                continue;
            }
            JsonObject args = new JsonObject();
            args.addProperty("action", "requestRouteInfo");
            args.addProperty("address", add);
            connList.get(add).sendMsg(JsonUtil.toJson(args));
        }
    }

    public List<OtherNCInfo> getInfo() {
        LOGGER.debug("getInfo: ");

        List<OtherNCInfo> list = new ArrayList<>();
        for (Map.Entry<String, CMNode> stringCMNodeEntry : NodeCenterActions.nodeInfos.entrySet()) {
            CMNode node = stringCMNodeEntry.getValue();
            if (null == node.contracts)
                continue;
            for (ContractDesp desp : node.contracts) {
                if (desp.type == ContractExecType.Sole || desp.getIsMaster()) {
                    list.add(new OtherNCInfo(desp.contractID, desp.contractName, node.masterAddress));
                }
            }
        }
        return list;
    }

    public void connectToNC(String target) throws InterruptedException {
        LOGGER.debug("connectToNC: ");
        if (target == null || !target.contains(":")) {
            LOGGER.error("connectToNC failed, target nc address is illegal! target=" + target);
            return;
        }

        final Bootstrap b = new Bootstrap();
        b.option(ChannelOption.CONNECT_TIMEOUT_MILLIS, 5000);
        EventLoopGroup group = new NioEventLoopGroup();
        b.group(group);
        NCConnector conn = new NCConnector(target, b, new NCClientHandler());
        connList.put(target, conn);
        b.channel(NioSocketChannel.class)
                .handler(
                        new ChannelInitializer<SocketChannel>() {
                            @Override
                            protected void initChannel(SocketChannel ch) {
                                ChannelPipeline p = ch.pipeline();
                                p.addLast(new DelimiterCodec()).addLast(conn.handler);
                            }
                        });


        if (!conn.handler.isOpen()) {
            String[] ipAndPort = target.split(":");
            b.connect(ipAndPort[0], Integer.parseInt(ipAndPort[1]))
                    .sync()
                    .channel();
        }

        reconnect(target);
    }

    public void reconnect(String target) throws InterruptedException {
        LOGGER.debug("reconnect: ");

        if (!connList.containsKey(target))
            return;

        NCConnector conn = connList.get(target);
        if (!conn.handler.isOpen()) {
            String[] ipAndPort = conn.ipAndPort.split(":");
            conn.bootstrap
                    .connect(ipAndPort[0], Integer.parseInt(ipAndPort[1]))
                    .sync()
                    .channel();
        }
    }

    static class NCConnector {
        Bootstrap bootstrap;
        NCClientHandler handler;
        String ipAndPort;

        public NCConnector(String s, Bootstrap b, NCClientHandler h) {
            bootstrap = b;
            handler = h;
            ipAndPort = s;
        }

        public void sendMsg(String s) {
            handler.sendMsg(s);
        }

        public void close() {
            handler.close();
            OtherNCProxy.instance.connList.remove(ipAndPort);
        }
    }
}
