package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.Jedion.KV;
import org.bdware.sc.bean.ContractDesp;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.action.Action;
import org.zz.gmhelper.SM2Util;

import java.io.File;
import java.util.*;

public class UnitActions {
    private static final Logger LOGGER = LogManager.getLogger(UnitActions.class);
    private final NCManagerAction managerAction;

    public UnitActions(NCManagerAction managerAction) {
        this.managerAction = managerAction;
    }
    private void simpleReply(ResultCallback resultCallback, String action, Object data){
        Map<String,Object> jsonResult = new HashMap<>();
        jsonResult.put("action", action);
        jsonResult.put("data", data);
        resultCallback.onResult(jsonResult);
    }
    private static String convertContractName(String contractID) {
        ContractDesp desp = NodeCenterActions.getContractByName(contractID);
        if (desp != null) return desp.contractID;
        return contractID;
    }

    /**
     * 显示当前用户可见的集群
     *
     * @param json
     * @param resultCallback
     */
    @Action(userPermission = 1 << 6, async = true)
    public void listTrustUnits(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        final String pubKey = managerAction.pubKey;
        List<KV> allunits = KeyValueDBUtil.instance.getKeyValues(NCTables.TrustUnitsDB.toString());
        if (pubKey != null
                && KeyValueDBUtil.instance
                .getValue(NCTables.ConfigDB.toString(), "__CenterManager__")
                .contains(pubKey)) {
            simpleReply(resultCallback,"onListTrustUnits",allunits);
            LOGGER.debug("is center manager");
            long end = System.currentTimeMillis();
            LOGGER.debug("[listTrustUnits:time]" + (end - start));
            return;
        }
        // 合约管理者获取自己创建的units
        Iterator<KV> it = allunits.iterator();
        List<KV> ciunits = new ArrayList<>();
        while (it.hasNext()) {
            KV kv = it.next();
            if (kv.key.contains(pubKey)) {
                ciunits.add(kv);
            }
        }
        simpleReply(resultCallback,"onListTrustUnits",ciunits);
        long end = System.currentTimeMillis();
        LOGGER.debug("[listTrustUnits:time]" + (end - start));
    }

    @Action(userPermission = 1 << 6, async = true)
    public void createTrustUnit(JsonObject json, ResultCallback resultCallback) {
        final String pubKey = managerAction.pubKey;
        LOGGER.debug("[createTrustUnit] " + json.get("data").toString());
        KeyValueDBUtil.instance.setValue(
                NCTables.TrustUnitsDB.toString(),
                pubKey + "_" + json.get("msg").getAsString(),
                json.get("data").toString());
        Map<String, Object> ret = new HashMap<>();
        ret.put("action", "onCreateTrustUnit");
        ret.put("status", "Success");
        resultCallback.onResult(ret);
    }

    @Action(userPermission = 1 << 6, async = true)
    public void deleteTrustUnit(JsonObject json, ResultCallback resultCallback) {
        //		final String pubKey = managerAction.pubKey;
        LOGGER.debug(
                "[deleteTrustUnit] before"
                        + KeyValueDBUtil.instance.getValue(
                        NCTables.TrustUnitsDB.toString(), json.get("data").getAsString()));
        KeyValueDBUtil.instance.delete(
                NCTables.TrustUnitsDB.toString(), json.get("data").getAsString());
        LOGGER.debug(
                "[deleteTrustUnit] after"
                        + KeyValueDBUtil.instance.getValue(
                        NCTables.TrustUnitsDB.toString(), json.get("data").getAsString()));
        Map<String, Object> ret = new HashMap<>();
        ret.put("action", "onDeleteTrustUnit");
        ret.put("status", "Success");
        resultCallback.onResult(ret);
    }

//    @Action(async = true)
//    public void executeContract(JsonObject args, final ResultCallback rc) {
//        executeContractTrustfully(args, rc);
//    }

    @Action(async = true, userPermission = 1L << 6)
    public void getNodeTrustUnits(JsonObject json, ResultCallback resultCallback) {
        String nodeID = json.get("data").getAsString();
        List<KV> allunits = KeyValueDBUtil.instance.getKeyValues(NCTables.TrustUnitsDB.toString());
        Iterator<KV> it = allunits.iterator();
        List<KV> nodeunits = new ArrayList<KV>();
        while (it.hasNext()) {
            KV kv = it.next();
            if (kv.value.contains(nodeID)) {
                nodeunits.add(kv);
            }
        }
        simpleReply(resultCallback,"onGetNodeTrustUnits",nodeunits);
    }

//    @Action(async = true, userPermission = 0)
//    public void executeContractTrustfully(JsonObject args, final ResultCallback rc) {
//        String contractID = args.get("contractID").getAsString();
//
//        contractID = convertContractName(contractID);
//
//        args.remove("contractID");
//        args.addProperty("contractID", contractID);
//        String requestID;
//        if (!args.has("requestID")) {
//            requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
//            args.addProperty("requestID", requestID);
//        } else requestID = args.get("requestID").getAsString();
//        args.remove("action");
//        args.addProperty("action", "executeContractTrustfully");
//        MultiPointContractInfo mInfo = NodeCenterActions.contractID2Members.get(contractID);
//        args.addProperty("seq", mInfo.getNextSeq());
//        args.addProperty("needSeq", ContractType.needSeq(mInfo.type));
//        List<String> nodes = mInfo.members;
//        logger.debug("tid:" + Thread.currentThread().getId() + " contractID:" + contractID);
//        if (nodes == null) {
//            rc.onResult(
//                    "{\"status\":\"Error\",\"result\":\"can't locate contract\",\"action\":\"onExecuteContractTrustfully\"}");
//        } else {
//            mInfo.rcf.execute(requestID, rc, JsonUtil.toJson(args));
//        }
//
//        // add request cache,only for requestAll type contract
//        if (ContractType.needSeq(mInfo.type)) {
//            RequestCache reqc = NodeCenterActions.getCache(contractID);
//            reqc.put(args.get("seq").getAsString(), args);
//        }
//    }


    //only for test ADSP
//    public static void testExecuteContract(JsonObject args, final ResultCallback rc) {
//        String contractID = args.get("contractID").getAsString();
//
//        contractID = convertContractName(contractID);
//
//        args.remove("contractID");
//        args.addProperty("contractID", contractID);
//        String requestID;
//        if (!args.has("requestID")) {
//            requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
//            args.addProperty("requestID", requestID);
//        } else requestID = args.get("requestID").getAsString();
//        args.remove("action");
//        args.addProperty("action", "executeContractTrustfully");
//        MultiPointContractInfo mInfo = NodeCenterActions.contractID2Members.get(contractID);
//        System.out.println("mInfo是空吗？" + mInfo == null);
//
//        args.addProperty("seq", mInfo.getNextSeq());
//        args.addProperty("needSeq", ContractType.needSeq(mInfo.type));
//        List<String> nodes = mInfo.members;
//        logger.debug("tid:" + Thread.currentThread().getId() + " contractID:" + contractID);
//        if (nodes == null) {
//            rc.onResult(
//                    "{\"status\":\"Error\",\"result\":\"can't locate contract\",\"action\":\"onExecuteContractTrustfully\"}");
//        } else {
//            mInfo.rcf.execute(requestID, rc, JsonUtil.toJson(args));
//        }
//
//        //add request cache,only for requestAll type contract
//        if(ContractType.needSeq(mInfo.type)){
//            RequestCache reqc = NodeCenterActions.getCache(contractID);
//            reqc.put(args.get("seq").getAsString(),args);
//        }
//    }

    @Action(async = true, userPermission = 1L << 6)
    public void listContractProcess(JsonObject args, final ResultCallback rc) {
        List<ContractDesp> info = new ArrayList<>();
        LOGGER.debug(
                "[contracts] "
                        + JsonUtil.toPrettyJson(NodeCenterActions.nodeInfos));
        LOGGER.debug(
                "[cid2Nodes]"
                        + JsonUtil.toPrettyJson(NodeCenterActions.contractID2Members));
        for (String key : NodeCenterActions.contractID2Members.keySet()) {
            ContractDesp desp = NodeCenterActions.getContractByID(key);
            if (desp != null) info.add(desp);
        }
        simpleReply(rc,"onListContractProcess",JsonUtil.toJson(info));
    }

    @Action(userPermission = 1 << 6, async = true)
    public void listMultiPointContractProcess(JsonObject json, ResultCallback resultCallback) {
        List<ContractDesp> info = new ArrayList<>();
        LOGGER.debug(
                "[contracts] "
                        + JsonUtil.toPrettyJson(NodeCenterActions.nodeInfos));
        LOGGER.debug(
                "[cid2Nodes]"
                        + JsonUtil.toPrettyJson(NodeCenterActions.contractID2Members));
        for (String key : NodeCenterActions.contractID2Members.keySet()) {
            ContractDesp desp = NodeCenterActions.getContractByID(key);
            if (desp != null) info.add(desp);
        }
        simpleReply(resultCallback,"onListMultiPointContractProcess",JsonUtil.toJson(info));
    }

    @Action(userPermission = 1 << 6, async = true)
    public void stopMultiPointContractProcess(JsonObject json, ResultCallback resultCallback) {
        String contractID = json.get("contractID").getAsString();
        MultiPointContractInfo info = NodeCenterActions.contractID2Members.get(contractID);
        NodeCenterActions.contractID2Members.remove(contractID);
        String req = "{\"action\":\"killContractProcess\",\"contractID\":\"" + contractID + "\"}";
        for (String member : info.members) {
            CMNode node;
            node = NodeCenterActions.nodeInfos.get(member);
            node.connection.controller.sendMsg(req);
        }
        listMultiPointContractProcess(json, resultCallback);
    }

    @Action(async = true)
    public void distributeYPK(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("[UnitActions] distributeYPK :  -----------position1");

        String pubKey = args.get("pubKey").getAsString();
        String signature = args.get("signature").getAsString();
        String projectName = args.get("projectName").getAsString();
        String[] strs = args.get("nodeIDs").getAsString().split(",");
        Set<String> nodePubKeys = new HashSet<>(); // nodes' pubkey
        for (String str : strs) {
            if (str != null && str.length() > 0) nodePubKeys.add(str);
        }
        Map<String, String> nodes = new HashMap<>(); // node's pubKey-node'a name
        for (CMNode node : NodeCenterActions.nodeInfos.values()) {
            if (nodePubKeys.contains(node.pubKey)) {
                nodes.put(node.pubKey, node.nodeName);
                LOGGER.debug("nodes add " + node.pubKey + "   " + node.nodeName);
            }
        }

        boolean result =
                SM2Util.plainStrVerify(
                        pubKey, "DistributeYPK|" + projectName + "|" + pubKey, signature);

        LOGGER.debug("[UnitAcitons] 验签:" + result + " -> projectName:" + projectName);

        String ypkPath = "./NodeCenterDB/NC_YPKs/";
        File dir = new File(ypkPath);
        if (!dir.exists()) {
            dir.mkdirs();
        }

        String filePath = new File(dir, projectName).getAbsolutePath();
        File f = new File(filePath);
        String fileName = null;
        fileName = f.getName();
        LOGGER.debug("[UnitActions] distributeYPK : fileName=" + fileName);
        LOGGER.debug("[UnitActions] nodeNames: " + JsonUtil.toJson(nodes));

        String isPrivate = "true";
        for (String ID : nodes.keySet()) {
            CMNode node = NodeCenterActions.nodeInfos.get(ID);
            node.connection.sendProject(filePath, isPrivate, pubKey, rc);
        }
    }

//    @Action(async = true, userPermission = 1L << 6)
//    public void startContractMultiPoint(JsonObject args, final ResultCallback rc) {
//        Map<String, String> request = new HashMap<String, String>();
//        Contract contract = new Contract();
//        if (args.has("type")) {
//            contract.setType(ContractType.valueOf(args.get("type").getAsString()));
//        }
//
//        // this multiPointContractInfo problem
//        MultiPointContractInfo multiPointContractInfo = new MultiPointContractInfo();
//        multiPointContractInfo.type = contract.getType();
//
//        logger.debug("startContractMultiPoint:");
//        logger.debug(JsonUtil.toPrettyJson(args));
//        if (args.has("script")) contract.setScript(args.get("script").getAsString());
//        else if (args.has("path")) contract.setScript(args.get("path").getAsString());
//        else if (args.has("projectName"))
//            contract.setScript("/" + args.get("projectName").getAsString());
//        // contract.setScript("/" + args.get("projectName").getAsString() + "/manifest.json");
//
//        SM2 sm2 = new SM2();
//        SM2KeyPair sm2Key = sm2.generateKeyPair();
//        contract.setID(sm2Key.getPublicKeyStr().hashCode() + "");
//        contract.setKey(sm2Key.getPrivateKey().toString(16));
//        contract.setPublicKey(sm2Key.getPublicKeyStr());
//
//        contract.setNodeCenterRepoDOI(DOIPMainServer4NC.repoIdentifier);
//        // 多点合约部署时，由NC预先注册一个DOI以备使用
//        if (DoConfig.openContractDORegister) {
//            try {
//                HandleService hs = new HandleService(HandleServiceUtils.hrRegister);
//                String tmpDOI = hs.registerContract(DOIPMainServer4NC.repoIdentifier);
//                if (tmpDOI == "" || tmpDOI == null)
//                    contract.setDOI("RegisterFailed");
//                else
//                    contract.setDOI(tmpDOI);
//            }
//            catch (Exception e) {
//                e.printStackTrace();
//                contract.setDOI("RegisterFailed");
//            }
//        }
//
//        String[] nodeNames =
//                args.get("peersID").getAsString().split(","); // all nodes' peerID in the unit
//        Set<String> nodeNameSet = new HashSet<>(); // nodes' peerID
//        for (String str : nodeNames) {
//            if (str != null && str.length() > 0) nodeNameSet.add(str);
//        }
//
//        List<String> udpids = new ArrayList<>(); // nodes' udpID in the unit
//
//        for (CMNode node : NodeCenterActions.nodeinfos.values()) {
//            if (nodeNameSet.contains(node.nodeName)) {
//                udpids.add(node.udpID);
//            }
//        }
//        final long curr = System.currentTimeMillis();
//        String requestID = curr + "_" + (int) (Math.random() * 10000);
//        NodeCenterActions.ResultCollector collector =
//                new NodeCenterActions.ResultCollector(
//                        requestID,
//                        new ResultCallback() {
//                            @Override
//                            public void onResult(String str) {
//                                Map<String, String> ret = new HashMap<>();
//                                ret.put("action", "onStartContractTrustfullyResult");
//                                ret.put("data", str);
//                                ret.put("executionTime", (System.currentTimeMillis() - curr) + "");
//                                rc.onResult(JsonUtil.toJson(ret));
//                            }
//                        },
//                        udpids.size());
//        NodeCenterActions.sync.sleepWithTimeout(requestID, collector, 20);
//
//        //		ArrayList<String> udpids=gson.fromJson(args.get("members"),
//        //                new TypeToken<ArrayList<String>>(){}.getType());
//        Map<String, String> members = new HashMap<>(); // nodes' <udpID,udpAddress>
//        for (int i = 0; i < udpids.size(); i++) {
//            members.put(
//                    udpids.get(i),
//                    NodeCenterActions.runner.getUDPAddr(Integer.valueOf(udpids.get(i)))
//                            + ":"
//                            + (i == 0)); // regard the first node as the master in the unit
//        }
//
//        List<String> nodes = new ArrayList<>(); // nodes' pubKey
//        for (CMNode node : NodeCenterActions.nodeinfos.values()) {
//            if (nodeNameSet.contains(node.nodeName)) {
//                nodes.add(node.pubKey);
//            }
//        }
//
//        request.put("memberStr", JsonUtil.toJson(members));
//        request.put("isPrivate", args.get("isPrivate").getAsString());
//        request.put("pubKey", managerAction.pubKey);
//        request.put("action", "startContractTrustfully");
//        request.put("requestID", requestID);
//        multiPointContractInfo.members = nodes;
//
//        contract.setNumOfCopies(1);
//        switch (multiPointContractInfo.type) {
//            case Sole:
//                logger.debug("Can't support Solo in multi-point mode");
//                return;
//            case RequestOnce:
//                multiPointContractInfo.rcf = new RequestOnceExecutor(multiPointContractInfo);
//                break;
//            case ResponseOnce:
//                multiPointContractInfo.rcf = new ResponseOnceExecutor(multiPointContractInfo);
//                break;
//            case RequestAllResponseFirst:
//                multiPointContractInfo.rcf = new RequestAllExecutor(multiPointContractInfo, 1);
//                contract.setNumOfCopies(nodes.size());
//                break;
//            case RequestAllResponseHalf:
//                multiPointContractInfo.rcf =
//                        new RequestAllExecutor(multiPointContractInfo, nodes.size() / 2 + 1);
//                contract.setNumOfCopies(nodes.size());
//                break;
//            case RequestAllResponseAll:
//                multiPointContractInfo.rcf =
//                        new RequestAllExecutor(multiPointContractInfo, nodes.size());
//                contract.setNumOfCopies(nodes.size());
//                break;
//        }
//        NodeCenterActions.contractID2Members.put(contract.getID(), multiPointContractInfo);
//
//        request.put("contractStr", JsonUtil.toJson(contract));
//        String startReq = JsonUtil.toJson(request);
//        logger.debug("启动合约:" + startReq);
//
//        for (String nodeID : nodes) {
//            CMNode node;
//            node = NodeCenterActions.nodeinfos.get(nodeID);
//            node.connection.controller.sendMsg(startReq);
//
//            if (!NodeCenterActions.recoverMap.containsKey(nodeID)) {
//                NodeCenterActions.recoverMap.put(nodeID, new ConcurrentHashMap<>());
//            }
//            ContractRecord record = new ContractRecord(contract.getID());
//            record.members = new HashMap<>(members);
//            NodeCenterActions.recoverMap.get(nodeID).put(contract.getID(), record);
//        }
//
//        rc.onResult(
//                "{\"status\":\"Success\",\"result\":\""
//                        + contract.getID()
//                        + "\",\"action\":\"onStartTrustfulContract\"}");
//    }
}
